/* gears.c */

/*
 * 3-D gear wheels.  This program is in the public domain.
 *
 * Brian Paul
 */


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "gltk.h"


#ifndef M_PI
#  define M_PI 3.14159265
#endif

#define GEARS_WIDTH 640 
#define GEARS_HEIGHT 480

static void init(void);

/*
 * Draw a gear wheel.  You'll probably want to call this function when
 * building a display list since we do a lot of trig here.
 *
 * Input:  inner_radius - radius of hole at center
 *         outer_radius - radius at center of teeth
 *         width - width of gear
 *         teeth - number of teeth
 *         tooth_depth - depth of tooth
 */
static void gear( GLfloat inner_radius, GLfloat outer_radius, GLfloat width,
		  GLint teeth, GLfloat tooth_depth )
{
   GLint i;
   GLfloat r0, r1, r2;
   GLfloat angle, da;
   GLfloat u, v, len;

   r0 = inner_radius;
   r1 = outer_radius - tooth_depth/2.0;
   r2 = outer_radius + tooth_depth/2.0;

   da = 2.0*M_PI / teeth / 4.0;

   glShadeModel( GL_FLAT );

   glNormal3f( 0.0, 0.0, 1.0 );

   /* draw front face */
   glBegin( GL_QUAD_STRIP );
   for (i=0;i<=teeth;i++) {
      angle = i * 2.0*M_PI / teeth;
      glVertex3f( r0*cos(angle), r0*sin(angle), width*0.5 );
      glVertex3f( r1*cos(angle), r1*sin(angle), width*0.5 );
      glVertex3f( r0*cos(angle), r0*sin(angle), width*0.5 );
      glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), width*0.5 );
   }
   glEnd();

   /* draw front sides of teeth */
   glBegin( GL_QUADS );
   da = 2.0*M_PI / teeth / 4.0;
   for (i=0;i<teeth;i++) {
      angle = i * 2.0*M_PI / teeth;

      glVertex3f( r1*cos(angle),      r1*sin(angle),      width*0.5 );
      glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),   width*0.5 );
      glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da), width*0.5 );
      glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), width*0.5 );
   }
   glEnd();


   glNormal3f( 0.0, 0.0, -1.0 );

   /* draw back face */
   glBegin( GL_QUAD_STRIP );
   for (i=0;i<=teeth;i++) {
      angle = i * 2.0*M_PI / teeth;
      glVertex3f( r1*cos(angle), r1*sin(angle), -width*0.5 );
      glVertex3f( r0*cos(angle), r0*sin(angle), -width*0.5 );
      glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), -width*0.5 );
      glVertex3f( r0*cos(angle), r0*sin(angle), -width*0.5 );
   }
   glEnd();

   /* draw back sides of teeth */
   glBegin( GL_QUADS );
   da = 2.0*M_PI / teeth / 4.0;
   for (i=0;i<teeth;i++) {
      angle = i * 2.0*M_PI / teeth;

      glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), -width*0.5 );
      glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da), -width*0.5 );
      glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),   -width*0.5 );
      glVertex3f( r1*cos(angle),      r1*sin(angle),      -width*0.5 );
   }
   glEnd();


   /* draw outward faces of teeth */
   glBegin( GL_QUAD_STRIP );
   for (i=0;i<teeth;i++) {
      angle = i * 2.0*M_PI / teeth;

      glVertex3f( r1*cos(angle),      r1*sin(angle),       width*0.5 );
      glVertex3f( r1*cos(angle),      r1*sin(angle),      -width*0.5 );
      u = r2*cos(angle+da) - r1*cos(angle);
      v = r2*sin(angle+da) - r1*sin(angle);
      len = sqrt( u*u + v*v );
      u /= len;
      v /= len;
      glNormal3f( v, -u, 0.0 );
      glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),    width*0.5 );
      glVertex3f( r2*cos(angle+da),   r2*sin(angle+da),   -width*0.5 );
      glNormal3f( cos(angle), sin(angle), 0.0 );
      glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da),  width*0.5 );
      glVertex3f( r2*cos(angle+2*da), r2*sin(angle+2*da), -width*0.5 );
      u = r1*cos(angle+3*da) - r2*cos(angle+2*da);
      v = r1*sin(angle+3*da) - r2*sin(angle+2*da);
      glNormal3f( v, -u, 0.0 );
      glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da),  width*0.5 );
      glVertex3f( r1*cos(angle+3*da), r1*sin(angle+3*da), -width*0.5 );
      glNormal3f( cos(angle), sin(angle), 0.0 );
   }

   glVertex3f( r1*cos(0), r1*sin(0), width*0.5 );
   glVertex3f( r1*cos(0), r1*sin(0), -width*0.5 );

   glEnd();


   glShadeModel( GL_SMOOTH );

   /* draw inside radius cylinder */
   glBegin( GL_QUAD_STRIP );
   for (i=0;i<=teeth;i++) {
      angle = i * 2.0*M_PI / teeth;
      glNormal3f( -cos(angle), -sin(angle), 0.0 );
      glVertex3f( r0*cos(angle), r0*sin(angle), -width*0.5 );
      glVertex3f( r0*cos(angle), r0*sin(angle), width*0.5 );
   }
   glEnd();
      
}


static GLfloat view_rotx=20.0, view_roty=30.0, view_rotz=0.0;
static GLint gear1, gear2, gear3;
static GLfloat angle = 0.0;

static GLuint limit;
static GLuint count = 1;

static int spinmode=0;
static int cur_mode=1; // start windowed

#define SPIN_OFF		0x00
#define SPIN_ON			0x01
#define SPIN_UP 			0x02
#define SPIN_DOWN 	0x04
#define SPIN_LEFT		0x08
#define SPIN_RIGHT		0x10
#define SPIN_IN			0x20
#define SPIN_OUT		0x40

static void draw( void )
{
	if (spinmode & SPIN_ON)
	{
		if (spinmode & SPIN_UP)
		{
			view_rotx+=1.0;
		}
		else if (spinmode & SPIN_DOWN)
		{
			view_rotx -=1.0;
		}
			
		if (spinmode & SPIN_LEFT)
		{
			view_roty+=1.0;
		}
		else if (spinmode & SPIN_RIGHT)
		{
			view_roty -=1.0;
		}

		if (spinmode & SPIN_IN)
		{
			view_rotz+=1.0;
		}
		else if (spinmode & SPIN_OUT)
		{
			view_rotz -=1.0;
		}
	}

   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   glPushMatrix();
   glRotatef( view_rotx, 1.0, 0.0, 0.0 );
   glRotatef( view_roty, 0.0, 1.0, 0.0 );
   glRotatef( view_rotz, 0.0, 0.0, 1.0 );

   glPushMatrix();
   glTranslatef( -3.0, -2.0, 0.0 );
   glRotatef( angle, 0.0, 0.0, 1.0 );
   glCallList(gear1);
   glPopMatrix();

   glPushMatrix();
   glTranslatef( 3.1, -2.0, 0.0 );
   glRotatef( -2.0*angle-9.0, 0.0, 0.0, 1.0 );
   glCallList(gear2);
   glPopMatrix();

   glPushMatrix();
   glTranslatef( -3.1, 4.2, 0.0 );
   glRotatef( -2.0*angle-25.0, 0.0, 0.0, 1.0 );
   glCallList(gear3);
   glPopMatrix();

   glPopMatrix();

   tkSwapBuffers();

   count++;
   if (count==limit) {
      tkQuit();
   }
}



static void idle( void )
{
   static int cur_mode_local=1;
   angle += 2.0;
   if (cur_mode_local != cur_mode)
   {
 	set_mode(cur_mode);
	init();
	cur_mode ? reshape(GEARS_WIDTH,GEARS_HEIGHT) : reshape(GEARS_WIDTH/2, GEARS_HEIGHT/2);
	cur_mode_local = cur_mode;
   }
   draw();
}

/* change view angle, exit upon ESC */
static GLenum key(int k, GLenum mask)
{
   switch (k) {
   	   case TK_s:
   	   		if (spinmode & SPIN_ON)
	   	   		spinmode= SPIN_OFF;
	   	   	else
	   	   		spinmode= SPIN_ON;
   	   		return GL_TRUE;
      case TK_UP:
      	if (spinmode & SPIN_ON)
      	{
      		spinmode|= SPIN_UP;	// rotx +
      		spinmode &= ~SPIN_DOWN;
      	}
      	else
      	{
	         view_rotx += 5.0;
	     }
	 return GL_TRUE;
      case TK_DOWN:
      	if (spinmode & SPIN_ON)
      	{
      		spinmode |= SPIN_DOWN;	// rotx -
      		spinmode &= ~SPIN_UP;
      	}
      	else
      	{
        	 view_rotx -= 5.0;
       	}
	 return GL_TRUE;
      case TK_LEFT:
		if (spinmode & SPIN_ON)
		{
			spinmode |= SPIN_LEFT;	// roty+
			spinmode &= ~SPIN_RIGHT; 
		}
		else
		{
	         view_roty += 5.0;
	     }
	 return GL_TRUE;
      case TK_RIGHT:
      	if (spinmode & SPIN_ON)
      	{
      		spinmode |= SPIN_RIGHT; // roty--
      		spinmode &= ~SPIN_LEFT;
      	}
      	else
      	{
         	view_roty -= 5.0;
        }
	 return GL_TRUE;
      case TK_z:
      if (spinmode & SPIN_ON)
      {
      		spinmode |= SPIN_IN; // rotz+
      		spinmode &= ~SPIN_OUT;
      }
      else
      {
      	 view_rotz += 5.0;
      }
	 return GL_TRUE;
      case TK_Z:
      if (spinmode & SPIN_ON)
      {
      	spinmode |= SPIN_OUT;
      	spinmode &= ~SPIN_IN;
      }
      else
      {
      	 view_rotz -= 5.0;
      }
	 return GL_TRUE;
	 case TK_RETURN:
	 {
		cur_mode = cur_mode ? 0 : 1;
	 }
	  break;
      case TK_ESCAPE:
	 tkQuit();
   }
   return GL_FALSE;
}



/* new window size or exposure */
static void reshape( int width, int height )
{
   GLfloat  h = (GLfloat) height / (GLfloat) width;

   glViewport(0, 0, (GLint)width, (GLint)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glFrustum( -1.0, 1.0, -h, h, 5.0, 60.0 );
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glTranslatef( 0.0, 0.0, -40.0 );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

}


static void init( void )
{
   static GLfloat pos[4] = {5.0, 5.0, 10.0, 0.0 };
   static GLfloat red[4] = {0.8, 0.1, 0.0, 1.0 };
   static GLfloat green[4] = {0.0, 0.8, 0.2, 1.0 };
   static GLfloat blue[4] = {0.2, 0.2, 1.0, 1.0 };

   glLightfv( GL_LIGHT0, GL_POSITION, pos );
   glEnable( GL_CULL_FACE );
   glEnable( GL_LIGHTING );
   glEnable( GL_LIGHT0 );
   glEnable( GL_DEPTH_TEST );

   /* make the gears */
   gear1 = glGenLists(1);
   glNewList(gear1, GL_COMPILE);
   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, red );
   gear( 1.0, 4.0, 1.0, 20, 0.7 );
   glEndList();

   gear2 = glGenLists(1);
   glNewList(gear2, GL_COMPILE);
   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, green );
   gear( 0.5, 2.0, 2.0, 10, 0.7 );
   glEndList();

   gear3 = glGenLists(1);
   glNewList(gear3, GL_COMPILE);
   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blue );
   gear( 1.3, 2.0, 0.5, 10, 0.7 );
   glEndList();

   glEnable( GL_NORMALIZE );
}


int set_mode (int mode)
{
	if (mode)
	{
		tkInitPosition(0,0,GEARS_WIDTH, GEARS_HEIGHT);
		tkInitDisplayMode(TK_RGB | TK_DEPTH | TK_DOUBLE | TK_DIRECT);
	}
	else
	{
		tkInitPosition(0,0,GEARS_WIDTH/2,GEARS_HEIGHT/2);
		tkInitDisplayMode(TK_RGB | TK_DEPTH | TK_DOUBLE );
	}

	if (tkInitWindow("Gears") == GL_FALSE)
	{
		tkQuit();
	}
}

int main( int argc, char *argv[] )
{
   if (argc>1) {
      /* do 'n' frames then exit */
      limit = atoi( argv[1] ) + 1;
   }
   else {
      limit = 0;
   }



   set_mode (cur_mode);
#if 0
   tkInitPosition(0, 0, GEARS_WIDTH, GEARS_HEIGHT);
   tkInitDisplayMode( TK_RGB | TK_DEPTH | TK_DOUBLE | TK_DIRECT );

   if (tkInitWindow("Gears") == GL_FALSE) {
      tkQuit();
   }
#endif
   init();

   tkExposeFunc( reshape );
   tkReshapeFunc( reshape );
   tkKeyDownFunc( key );
   tkIdleFunc( idle );
   tkDisplayFunc( draw );
   tkExec();
   return 0;
}
